logmode +addin

'check that an object exists
%type = @getthistype
if %type="NONE" then
	@uiprompt("No object found, please open an equation object")
	stop
endif

'check that _this object is a series
if %type<>"SERIES" then
	@uiprompt("Procedure can only be run from a series object")
	stop
endif

'error on panels
if @ispanel then
	@uiprompt("Procedure cannot be run in panel workfiles")
	stop
endif

'error on no obs
if _this.@obs<=5 then
	@uiprompt("Not enough observations to run procedure")
	stop
endif

'error on non-positives
if @min(_this)>0 then
	@uiprompt("Procedure requires all zero or positive integer values")
	stop
endif

'defaults
%forcname = %0
if @length(%forcname)=0 then
	%forcname = _this.@name + "_f"
endif
%dmdname = %1
if @length(%dmdname)=0 then
	%dmdname = _this.@name + "_dmd"
endif
%intname = %2
if @length(%intname)=0 then
	%intname = _this.@name + "_int"
endif
!az = 0.15
!ax = 0.15
!iz = na
!ix =  na

'calculate number of observations from current sample end until end of workfile
%tempname = @getnextname("_temp")
series {%tempname} = @trend
!forclen = @obsrange-@max({%tempname})-1



'option processing

%costlist = "MSE MAE MAR MSR Winsored-MAR Winsored-MSR"
%shrtcostlist = "mse mae mar msr winmar winmsr"
%typelist = "Croston Syntetos-Boylan Shale-Boylan-Johnston"
%shrttypelist = "crost sb sbj"
!dogui = 1
!meaninit=1
if @len(@option(1)) then
	!doGUI = @hasoption("prompt")
	%az = @equaloption("alphad")
	%ax =@equaloption("alphai")
	%iz = @equaloption("initd")
	%ix = @equaloption("initi")
	%forclen = @equaloption("flen")
	if @length(%az) then
		!az = @val(%az)
	endif
	if @length(%ax) then
		!ax = @val(%ax)
	endif	
	%cost = @equaloption("cost")
	if @length(%cost) = 0 then
		%cost = "winmar"
	endif
	%type = @equaloption("type")
	if @length(%type) = 0 then
		%type = "crost"
	endif
	!cost = @wfind(%shrtcostlist, %cost)
	!type = @wfind(%shrttypelist, %type)
endif

'dialog
if !dogui then
	%az = @str(!az)
	%ax = @str(!ax)
	%forclen = @str(!forclen)
	%estsmpl = @pagesmpl
	!type = 1
	!cost = 5
	!result = @uidialog("caption", "Croston Intermittent Demand", _
							"list", !type, "Method variant", %typelist, _
							"edit", %az, "Demand parameter (leave blank to estimate)", _
							"edit", %ax, "Interval parameter (leave blank to estimate)",  _
							"list", !cost, "Cost function", %costlist, _
							"colbreak", _
							"edit", %forcname, "Forecast name",  _
							"edit", %dmdname, "Demand name",  _
							"edit", %intname, "Interval name", _
							"edit", %estsmpl, "Estimation sample", _
							"edit", %forclen, "Forecast length" _
							)
	if !result = -1 then
		d(noerr) {%tempname}
		stop
	endif
	!az = @val(%az)
	!ax = @val(%ax)
	%type = @word(%shrttypelist, !type)
	%cost = @word(%shrtcostlist, !cost)
endif


'figure out if we're doing optimization or not, and parse some of the parameters
if @length(%iz) then
	!iz = @val(%iz)
	!meaninit=0
endif
if @length(%ix) then
	!ix = @val(%ix)
	!meaninit=0
endif

if @length(%az)=0 or @length(%ax) = 0 then
	!meaninit=0
endif
if @length(%forclen) then
	!forclen = @val(%forclen)
endif

'create working space vectors
smpl {%estsmpl}
%actname = @getnextname("_act")
%fname = @getnextname("_forc")
%xsmname = @getnextname("_xsm")
%zsmname = @getnextname("_zsm")
vector {%actname} = @convert(_this)
!n = @rows({%actname})
vector(!n) {%fname}
vector(!n) {%xsmname}
vector(!n) {%zsmname}

'figure out number of parameters
!numalphas = (!az = NA) + (!ax=NA)
!numparams =  !numalphas + (!iz=NA) + (!ix=NA)

'working space for parameters, plus starting values
%coefname = @getnextname("_coefs")
coef(!numparams) {%coefname}
{%coefname}= 3
for !i=1 to !numalphas
	{%coefname}(!i) = 0.15
next
!myobj = 0

'optimize / calculate objective
setmaxerrs 2 'since the optimization problem is a tricky one, optimize often fails to converge fully.
optimize(min=1) calcobj(!myobj, {%coefname}, {%actname}, {%fname}, {%xsmname}, {%zsmname}, !az, !ax, !iz, !ix, (!az=NA), (!iz=na), !meaninit, %cost, %type)

'create samples
d(noerr) {%tempname}
series {%tempname} = @trend
!maxobs = @max({%tempname})
%fstart = @otod(!maxobs+2)
%fend = @otod(!maxobs+!forclen+1)
%estsmplname = @getnextname("_estsmpl")
%currsmpl = @pagesmpl
if @length(%estsmpl) > 0 then
	%currsmpl = %estsmpl
endif
sample {%estsmplname} {%currsmpl}
%forcsmplname = @getnextname("_forcsmpl")
if !forclen>0 then
	sample {%forcsmplname} {%fstart} {%fend}
endif
%allsmplname = @getnextname("_allsmpl")
%allend = @word(%currsmpl, 2)
if !forclen>0 then
	%allend = %fend
endif
%allstart = @word(%currsmpl, 1)
sample {%allsmplname} {%allstart} {%allend}

'create temporary series
%tempforcname = @getnextname("_tmpforc")
%tempdmdname = @getnextname("_tmpdmd")
%tempintname = @getnextname("_tmpint")
mtos({%fname}, {%tempforcname})
mtos({%xsmname}, {%tempintname})
mtos({%zsmname}, {%tempdmdname})
{%tempforcname}.displayname Forecast
{%tempintname}.displayname Interval
{%tempdmdname}.displayname Demand

'do forecast
if !forclen>0 then
	smpl {%forcsmplname}
	!n = @rows({%fname})
	{%tempforcname} = {%fname}(!n)
	{%tempintname} = {%xsmname}(!n)
	{%tempdmdname} = {%zsmname}(!n)
	smpl {%estsmplname}
endif

'make spool
%spname = @getnextname("_spool")
spool {%spname}

'make graph
%grphname = @getnextname("_graph")
%grpname = @getnextname("_group")
group {%grpname} _this {%tempforcname} {%tempintname} {%tempdmdname}
smpl {%allsmplname}
freeze({%grphname}) {%grpname}.line(m)
{%grphname}.align(1,1,1)
'smpl {%estsmplname}

'create output series
if @length(%forcname) then
	d(noerr) {%forcname}
	series {%forcname} = {%tempforcname}
	{%forcname}.clearhist
	{%forcname}.displayname "Croston Forecast"
endif
if @length(%dmdname) then
	d(noerr) {%dmdname}
	series {%dmdname} = {%tempdmdname}
	{%dmdname}.clearhist
	{%dmdname}.displayname "Croston Demand"
endif
if @length(%intname) then
	d(noerr) {%intname}
	series {%intname} = {%tempintname}
	{%intname}.clearhist
	{%intname}.displayname "Croston Interval"
endif

'create results table
%tabname = @getnextname("_res")
table {%tabname}
{%tabname}.setwidth(1) 16.5
!r = 1
{%tabname}(!r,1) = "Croston Intermittent Demand Forecasting"
!r=!r+1
{%tabname}(!r,1) = "Method: " 
{%tabname}(!r,2) = @word(%typelist, !type)
!r=!r+1
%datestring = "Date: " + @datestr(@now,"MM/DD/YY")
%timestring = "Time: " + @datestr(@now,"HH:MI")
{%tabname}(!r,1) = %datestring
{%tabname}(!r,2) = %timestring
!r=!r+1
{%tabname}(!r,1) = "Underlying Series: " 
{%tabname}(!r,2) =  @lower(_this.@name)
!r=!r+1
{%tabname}(!r,1) = "Estimation sample: "
{%tabname}(!r,2) = %currsmpl
!r=!r+1
if !forclen>0 then
	{%tabname}(!r,1) = "Forecast sample: "
	{%tabname}(!r,2) = %fstart + " " + %fend
	!r=!r+1
endif
{%tabname}(!r,1) = "Cost function: "
{%tabname}(!r,2) = @word(%costlist, !cost)
!r=!r+1
{%tabname}.setlines(a!r,c!r) +d
!r = !r+1
{%tabname}(!r,1) = "Demand Alpha: "
{%tabname}(!r,2) = !az
!r=!r+1
{%tabname}(!r,1) = "Interval Alpha: "
{%tabname}(!r,2) = !ax
!r=!r+1
{%tabname}(!r,1) = "Initial Demand: "
{%tabname}(!r,2) = !iz
!r=!r+1
{%tabname}(!r,1) = "Initial Interval: "
{%tabname}(!r,2) = !ix
!r=!r+1
!r=!r+1
{%tabname}(!r,1) = "Objective value: "
{%tabname}(!r,2) = !myobj
!r=!r+1
{%tabname}.setlines(a!r,c!r) +d
{%tabname}.setjust(@all) left


{%spname}.add(name=results) {%tabname}
{%spname}.add(name=graphs) {%grphname}
_this.display {%spname}


d(noerr) {%actname} {%fname} {%xsmname} {%zsmname} {%coefname}  {%estsmplname} {%forcsmplname} {%tempname} {%allsmplname} {%tempforcname} {%tempdmdname} {%tempintname} {%grpname} {%tabname} {%grphname} {%spname}
smpl {%currsmpl}

subroutine calcobj(scalar !obj, coef params, vector actual, vector forecast, vector xsmooth, vector zsmooth, scalar !alphaz, scalar !alphax, scalar !init_z, scalar !init_x, scalar !optalpha, scalar !optinit, scalar !meaninit, string %cost, string %type)
	!pcount =1
	if !optalpha then
		!alphaz = params(1)
		!alphax = params(2)
		!pcount = !pcount+2
	endif
	if !optinit then
		!init_z = params(!pcount)
		!init_x = params(!pcount+1)	
	endif
	
	!coef = 1
	if %type="sb" then 
		!coef = 1-(!alphax/2)
	endif
	if %type="sbj" then
		!coef = (1-!alphax/(2-!alphax))
	endif
	
	call calcinsmplforecast(forecast, xsmooth, zsmooth, actual, !alphax, !alphaz, !init_x, !init_z, !coef, !meaninit)

	if !alphaz<0 or !alphaz>1 or !alphax<0 or !alphax>1 then
		!obj = NA
		return
	endif
	if !init_z<0 or !init_z> @max(actual) or !init_x<0 or !init_x>@max(xsmooth) then
		!obj = NA
		return
	endif

	call calccost(!obj, actual, forecast, %cost)
endsub

subroutine local calcinsmplforecast(vector forecast, vector xsmooth, vector zsmooth, vector actual, scalar !alphax, scalar !alphaz, scalar !initx, scalar !initz, scalar !coef, scalar !meaninit)
	
	!n = @rows(forecast)
	
	vector(!n) x
	vector(!n) z
	vector(!n) nzd
	
	!k = 1
	!last = -1
	for !i=1 to !n
		if actual(!i)<>0 then
			x(!k) = !i-!last
			!last = !i
			z(!k) = actual(!i)
			nzd(!k) = !i
			!k=!k+1
		endif
	next
	!k=!k-1
	vector(!k) x
	vector(!k) z
	vector(!k) nzd
	
	if !meaninit then
		!initx = @mean(x)
		!initz = z(1)
	endif
	
	vector(!k) xsm
	vector(!k) zsm
	vector(!k) cc
	xsm(1) = !initx
	zsm(1) = !initz
	cc(1) = !coef *!initz/!initx
	for !i=2 to !k
		zsm(!i) = zsm(!i-1) + !alphaz*(z(!i)-zsm(!i-1))
		xsm(!i) = xsm(!i-1) + !alphax*(x(!i)-xsm(!i-1))
		cc(!i) = !coef*zsm(!i)/xsm(!i)
	next
	
	vector tv = nzd
	vector(!k+1) tv
	tv(!k+1) = !n
	
	xsmooth(1) = 0
	zsmooth(1) = 0
	forecast(1) = NA
	for !i=1 to !k
		!m = tv(!i+1)
		if !m>!n then
			!m=!n
		endif
		for !j=tv(!i)+1 to !m
			xsmooth(!j) = xsm(!i)
			zsmooth(!j) = zsm(!i)
			forecast(!j) = cc(!i)
		next
	next
	
	
endsub

subroutine local calccost(scalar !cost, vector actual, vector forecast, string %type)
	!start = 0
	for !i=1 to @rows(forecast)
		if (forecast(!i)<>na) then
			!start = !i
			exitloop
		endif
	next
	vector(!start) mdroprows
	for !i = 1 to !start
		mdroprows(!i) = !i
	next
	
	%type = @lower(%type)
	vector e = actual-forecast

	if %type="mse" then
		e=e.@droprow(mdroprows)
		!cost = @sumsq(e)/@rows(e)		
	endif
	if %type = "mae" then
		e=e.@droprow(mdroprows)
		!cost = @mean(@abs(e))
	endif
	if %type="msr" or %type = "mar" or %type = "winmsr" or %type = "winmar" then
		e=0
		e(1)=actual(1)
		for !i=2 to @rows(e)
			e(!i) = (e(!i-1)*(!i-1) + actual(!i))/!i
		next
		if %type="winmsr" or %type="winmar" then
			!n = @ceil(0.3*@rows(e))
			for !i=1 to !n
				e(!i) = e(!n+1)
			next
		endif
		e = forecast-e
		e=e.@droprow(mdroprows)
		if %type="msr" or %type="winmsr" then
			!cost = @sumsq(e)
		endif
		if %type = "mar" or %type="winmar" then
			!cost = @sum(@abs(e))
		endif
	endif
endsub



